/********************************************************
*	Main source file for my first gba proggy.	*
*	Please note that this should be considered	*
*	as an ARM/GBA experiment, which explains	*
*	why a lot of things are done in a somewhat	*
*	"strange" way ;-)				*
*	Oh, and it's nowhere optimized.			*
*							*
*	Eloist (eloist@gmx.de)				*
*							*
********************************************************/

// Gba.h has all the register definitions
#include "gba.h"
#include <math.h>

// This is needed for the degree to rad conversion
#define PI 3.141592654

// Data included by the assembler (DCB, INCBIN)
extern u8 font;
extern char text;

// Special memory offset of the palette men
u16 *palette_mem = (u16*)0x5000000;

// Simple dot matrix in a 3D space
typedef struct POINT {
	int x;
	int y;
	int z;	
}POINT;

// Prototypes
void draw_cube_back (int distance, int angle_x, int angle_y, int angle_z);
POINT rotate_point_x (POINT point, int angle);
POINT rotate_point_y (POINT point, int angle);
POINT rotate_point_z (POINT point, int angle);
void setup_cube (void);
void draw_picture_back (int x, int y, int width, int height, byte *picture, int buffer);
void textout (int x, int y, char *string, int count, int buffer);
void line(int a, int b, int c, int d, int col);
int sgn (int value);
int abs (int value);

// Just some memory where "0" is stored and a pointer to it
u32 zero=0;
u32 *zero_pointer = &zero;

// Base offsets of the cube's corners from the center
POINT cube_base_coords [8];

// The rotated corners are stored here
POINT cube_current_coords [8];

// Descripton of the cube's lines (stored as [LINE] [0=FROM;1=TO]
int cube_lines [12] [2];

// Sinus and Cosinus lookup tables
float SIN [361];
float COS [361];

// Screen coordinates
int screen_x, screen_x2;
int screen_y, screen_y2;

// stupid way to do this
int string_atribs [27] [2] = {{ 5, 95},{14, 50},{18, 30},
			      {15, 45},{22, 10},{17, 35},
			      {20, 15},{11, 65},{11, 65},
			      {11, 65},{22, 10},{18, 30},
			      {22, 10},{18, 30},{18, 30},
			      { 6, 90},{13, 55},{20, 20},
			      { 4,100},{ 7, 85},{ 7, 85},
			      {13, 55},{ 7, 85},{17, 35},
			      {15, 45},{13, 55},{ 5, 95}};

// C code entry
int C_entry()
{
	int angle = 45;
	float f, rad;
	u8 fader = 0;
	int string_num = 0;
	int fade_back = 0;
	int fade_add = 1;
	int text_offset = 0;
	char *string = &text;
			
	// Setup the palette
	palette_mem [0] = 0;
	palette_mem [1] = 0x3E0;
	palette_mem [2] = 0x7FFF;
	
	// Setup the lookup tables
	for (f=0; f<=361; f++) {
		rad = ((f/360)*2)*PI;
		SIN [(int)f] = sin(rad);	
		COS [(int)f] = cos(rad);
	}
	
	// Setup the cube's coords and lines
	setup_cube();
		
	// Finally, setup the display to mode 4
	REG_DISPCNT = 0x04;
	REG_DISPSTAT = 0x08;
	
	while (1) {
						
		draw_cube_back (170, angle, angle, angle);
		
		__asm {
				mov 	r0, #0x4000006
			scanline_wait:
				ldrh	r1, [r0]
				cmp	r1, #160
				bne 	scanline_wait
			
				mov 	r2, #0x4000000 
				ldrh 	r0, [r2]
				and 	r1, r0, #0xFFFFFFEF
				tst 	r0, #0x10
				orreq 	r1, r1, #0x10
				strh 	r1, [r2]
		
				mov	r2, #0x6000000
				ldrh 	r1, [0x4000000]
				tst 	r1, #0x10
				addeq 	r2, r2, #0xA000
				ldr	r1, [zero_pointer]
				mov	r3, #0x20D0
				mov	r4, #0x8500 
				mov	r0, #0x40000D4
				str	r1, [r0], #0x4
				str	r2, [r0], #0x4
				strh	r3, [r0], #0x2
				strh	r4, [r0]
				
				add	angle, angle, #2
				cmp 	angle, #360
				movhs   angle, #0
				add 	fader, fader, fade_add
		}
		
		if ((fader==0x1F) && !fade_back) {
			fade_add = -1;
			fade_back = 1;			
		}
		else if ((fader==0) && fade_back) {
			
			__asm {
					mov	r0, #0x6008340
					mov	r1, #0x6012340
					mov	r2, #0x4B0
				clr_rest:
					str	r4, [r0], #4
					str	r4, [r1], #4
					subs	r2, r2, #1
					bne	clr_rest
			}
			
			textout (string_atribs[string_num][1], 140, string, string_atribs[string_num][0], 0);
			textout (string_atribs[string_num][1], 140, string, string_atribs[string_num][0], 1);
			
			fade_add = 1;
			fade_back = 0;
			
			string_num++;
			
			text_offset += string_atribs [string_num-1] [0];
			
			if (string_num == 27) {
				string_num = 0;
				text_offset = 0;
			}
			
			string = &text + text_offset;
		}
		
		palette_mem [2] = (fader<<10)|(fader<<5)|fader;
	}
}

void draw_cube_back (int distance, int angle_x, int angle_y, int angle_z)
{
	int i;
	POINT temp, temp2;
	
	for (i=0; i<8; i++) {
		temp = rotate_point_x (cube_base_coords[i], angle_x);
		temp = rotate_point_y (temp, angle_y);
		cube_current_coords[i] = rotate_point_z (temp, angle_z);
	}
	
	for (i=0; i<12; i++) {
		temp  = cube_current_coords [cube_lines[i][0]];
		temp2 = cube_current_coords [cube_lines[i][1]];
		
		screen_x =  ((256*temp.x)/(temp.z-distance)) + 120;
       		screen_y =  ((256*temp.y)/(temp.z-distance)) + 70;
		
		screen_x2 =  ((256*temp2.x)/(temp2.z-distance)) + 120;
       		screen_y2 =  ((256*temp2.y)/(temp2.z-distance)) + 70;
		
		line (screen_x, screen_y, screen_x2, screen_y2, 1);
	}
}

POINT rotate_point_x (POINT point, int angle)
{
	int y,z;

	y = (COS[angle]*point.y) - (SIN[angle]*point.z);
	z = (SIN[angle]*point.y) + (COS[angle]*point.z);
	point.y = y;
	point.z = z;
	
	return point;	
}

POINT rotate_point_y (POINT point, int angle)
{
	int x,z;
	
	x = (COS[angle]*point.x) + (SIN[angle]*point.z);
	z = (-(SIN[angle]*point.x)) + (COS[angle]*point.z);
	point.x = x;
	point.z = z;
	
	return point;	
}

POINT rotate_point_z (POINT point, int angle)
{
	int x,y;
	
	x = (COS[angle]*point.x) - (SIN[angle]*point.y);
      	y = (SIN[angle]*point.x) + (COS[angle]*point.y);
	point.x = x;
	point.y = y;
	
	return point;
}

void line(int a, int b, int c, int d, int col) 
{
  	int i,u,s,v,d1x,d1y,d2x,d2y,m,n,temp;

  	u   = c-a;       
  	v   = d-b;      
  	d2x = d1x = sgn(u);
  	d1y = sgn(v);  
  	d2y = 0;
  	m   = abs(u); 
  	n   = abs(v);  

	if (m<=n) {     
    		d2x = 0;
   	 	d2y = d1y;  
    		temp = m;
    		m = n;
    		n = temp;
    	}

 	s = m >> 1;

  	for (i=0;i<m+1;i++) {
    		__asm {
			mov 	r0, #0x6000000
		
			ldrh 	r8, [0x4000000]
			tst 	r8, #0x10
			addeq 	r0, r0, #0xA000	
		
			mul 	r1, b, #0xF0
			add 	r0, r0, r1
			add 	r0, r0, a
		
			mov 	r2, col
			strb 	r2, [r0]
		}	
    		
    		s += n;      
    		
    		if (s >= m) {
      			s -= m;
      			a += d1x;
      			b += d1y;
    		}
    		else {
      			a += d2x;
      			b += d2y;
    		}
  	}
}

void draw_picture (int x, int y, int width, int height, byte *picture, int buffer)
{
	int i,n;
	u8 data;
	u32 offset = x + ((y-1)*240);
	s32 offset2 = -width;
	
	u8 *current_screen = (u8*)0x600A000;
	
	if (buffer&&(REG_DISPCNT & 0x10)) {
		current_screen -= 0xA000;
	}
	else if (!buffer && !(REG_DISPCNT & 0x10)){
		current_screen -= 0xA000;
	}

	for (i=0; i<height; i++) 
	{
		offset += 240;	
		offset2 += width;
		for (n=0; n<width; n++) 
		{
			if ((data=picture[offset2+n]))
				current_screen[offset+n] = data;		
		}
	}
}

void textout (int x, int y, char *string, int count, int buffer)
{
	int index;
	int x_offset=0;
	
	for (index=0; index<count; index++) {	
		draw_picture((x+x_offset), y, 10, 10, (&font+((string[index]-0x23)*100)), buffer);
		
		x_offset+=10;	
	}
}

void setup_cube (void)
{
  cube_base_coords [0].x = -25; cube_base_coords [0].y =  25; cube_base_coords [0].z = -25;
  cube_base_coords [1].x =  25; cube_base_coords [1].y =  25; cube_base_coords [1].z = -25;
  cube_base_coords [2].x =  25; cube_base_coords [2].y =  25; cube_base_coords [2].z =  25;
  cube_base_coords [3].x = -25; cube_base_coords [3].y =  25; cube_base_coords [3].z =  25;
  cube_base_coords [4].x = -25; cube_base_coords [4].y = -25; cube_base_coords [4].z = -25;
  cube_base_coords [5].x =  25; cube_base_coords [5].y = -25; cube_base_coords [5].z = -25;
  cube_base_coords [6].x =  25; cube_base_coords [6].y = -25; cube_base_coords [6].z =  25;
  cube_base_coords [7].x = -25; cube_base_coords [7].y = -25; cube_base_coords [7].z =  25;
  
  cube_lines [0]  [0] = 0; cube_lines [0]  [1] = 1;
  cube_lines [1]  [0] = 1; cube_lines [1]  [1] = 2;
  cube_lines [2]  [0] = 2; cube_lines [2]  [1] = 3;
  cube_lines [3]  [0] = 3; cube_lines [3]  [1] = 0;
  cube_lines [4]  [0] = 0; cube_lines [4]  [1] = 4;
  cube_lines [5]  [0] = 1; cube_lines [5]  [1] = 5;
  cube_lines [6]  [0] = 2; cube_lines [6]  [1] = 6;
  cube_lines [7]  [0] = 3; cube_lines [7]  [1] = 7;
  cube_lines [8]  [0] = 4; cube_lines [8]  [1] = 5;
  cube_lines [9]  [0] = 5; cube_lines [9]  [1] = 6;
  cube_lines [10] [0] = 6; cube_lines [10] [1] = 7;
  cube_lines [11] [0] = 7; cube_lines [11] [1] = 4;
} 

int sgn (int value)
{
	if (value<0)
		return -1;
	else
		return 1;
}

int abs (int value)
{
	if (value<0)
		return (value-(2*value));	
	else
		return value;
}